# Copyright (c) 2002 Cunningham & Cunningham, Inc.
# Released under the terms of the GNU General Public License version 2 or later.


module Fit;


  class ParseException < Exception

    attr_reader :error_offset

    def initialize(msg, offset)
      super(msg)
      @error_offset = offset
    end
  end

  class ParseHolder
    attr_accessor :leader, :tag, :body, :endtag, :trailer,  :parts
    attr_accessor :more

    def ParseHolder.create(tag, body, parts, more)
      p = new
      p.leader = "\n"
      p.tag = "<#{tag}>"
      p.body = body
      p.endtag = "</#{tag}>"
      p.trailer = ""
      p.parts = parts
      p.more = more
      p
    end


    def size
      more ? more.size + 1 : 1
    end


    def last
      more ? more.last : self
    end


    def leaf()
      parts ? parts.leaf() : self
    end

    def at(i, *rest)
      node = (i.zero? || more.nil?) ? self : more.at(i-1)
      rest.each do |j|
        node = node.parts.at(j)
      end
      node
    end

    def add_to_tag(text)
      @tag[-1,0] = text
    end

    def add_to_body(text)
      @body += text
    end


    def text
      Parse.unescape(Parse.unformat(body)).strip
    end

    def print(out)
      out.print(@leader)
      out.print(@tag)
      if parts
	parts.print(out)
      else
	out.print(@body)
      end
      out.print(@endtag)
      if more
	more.print(out)
      else
	out.print(@trailer)
      end
    end

  end

  ######################################################################

  class Parse < ParseHolder


    DEFAULT_TAGS = [ "table", "tr", "td" ]


    def initialize(text, tags = DEFAULT_TAGS, level = 0, offset = 0)
      tag = tags[level]
      tag_pattern = %r{(<#{tag}.*?>)(.*?)(</#{tag}.*?>)}im

      unless (match = tag_pattern.match(text))
        raise ParseException.new("Can't find tag: #{tag}", offset)
      end
      
      @leader = match.pre_match
      @tag    = match[1]
      @body   = match[2]
      @endtag = match[3]
      @trailer = match.post_match
      
      if level+1 < tags.size
        @parts = Parse.new(@body, tags, level+1, offset + leader.size + @tag.size)
        @body = nil
      end
      
      if @trailer =~ /<#{tag}/i
        @more = Parse.new(@trailer, tags, level, offset + match[0].size)
        @trailer = nil
      end
    end



    def Parse.unformat(s) 
      s = s.dup
      1 while s.gsub!(/<.*?>/, '')
      s
    end

    SPECIALS = {
      'lt'   => '<',
      'gt'   => '>',
      'amp'  => '&',
      'nbsp' => ' '
    }

    def Parse.unescape(s)
      s.gsub(/&([^&]*?);/) { SPECIALS[$1.downcase] || $1}
    end


  end
end
